/* security.cc: NT security functions

   Copyright 1997, 1998, 1999 Cygnus Solutions.

   Originaly written by Gunther Ebert, gunther.ebert@ixos-leipzig.de
   Extensions by Corinna Vinschen <corinna.vinschen@cityweb.de>

This file is part of Cygwin.

This software is a copyrighted work licensed under the terms of the
Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
details. */

#include <grp.h>
#include <pwd.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "winsup.h"

extern BOOL allow_ntea;
BOOL allow_ntsec = FALSE;

SID_IDENTIFIER_AUTHORITY sid_auth[] = {
        {SECURITY_NULL_SID_AUTHORITY},
        {SECURITY_WORLD_SID_AUTHORITY},
        {SECURITY_LOCAL_SID_AUTHORITY},
        {SECURITY_CREATOR_SID_AUTHORITY},
        {SECURITY_NON_UNIQUE_AUTHORITY},
        {SECURITY_NT_AUTHORITY}
};

PSID
get_sid (PSID psid, DWORD s, DWORD cnt, DWORD *r)
{
  DWORD i;

  if (! psid || s > 5 || cnt < 1 || cnt > 8)
    return NULL;

  InitializeSid(psid, &sid_auth[s], cnt);
  for (i = 0; i < cnt; ++i)
    memcpy ((char *) psid + 8 + sizeof (DWORD) * i, &r[i], sizeof (DWORD));
  return psid;
}

PSID
get_ssid (PSID psid, const char *sid_str)
{
  char sid_buf[256];
  char *t;
  DWORD cnt = 0;
  DWORD s = 0;
  DWORD i, r[8];

  if (! sid_str || strncmp (sid_str, "S-1-", 4))
    return NULL;

  strcpy (sid_buf, sid_str);

  for (t = sid_buf + 4, i = 0; cnt < 8 && (t = strtok (t, "-")); t = NULL, ++i)
    if (i == 0)
      s = strtoul (t, NULL, 10);
    else
      r[cnt++] = strtoul (t, NULL, 10);

  return get_sid (psid, s, cnt, r);
}

BOOL
get_pw_sid (PSID sid, struct passwd *pw)
{
  char *sp = strrchr (pw->pw_gecos, ',');

  if (!sp)
    return FALSE;
  return get_ssid (sid, ++sp) != NULL;
}

BOOL
get_gr_sid (PSID sid, struct group *gr)
{
  return get_ssid (sid, gr->gr_passwd) != NULL;
}

PSID
get_admin_sid ()
{
  static NO_COPY char admin_sid_buf[40];
  static NO_COPY PSID admin_sid = NULL;

  if (!admin_sid)
    {
      admin_sid = (PSID) admin_sid_buf;
      get_ssid (admin_sid, "S-1-5-32-544");
    }
  return admin_sid;
}

PSID
get_system_sid ()
{
  static NO_COPY char system_sid_buf[40];
  static NO_COPY PSID system_sid = NULL;

  if (!system_sid)
    {
      system_sid = (PSID) system_sid_buf;
      get_ssid (system_sid, "S-1-5-18");
    }
  return system_sid;
}

PSID
get_creator_owner_sid ()
{
  static NO_COPY char owner_sid_buf[40];
  static NO_COPY PSID owner_sid = NULL;

  if (!owner_sid)
    {
      owner_sid = (PSID) owner_sid_buf;
      get_ssid (owner_sid, "S-1-3-0");
    }
  return owner_sid;
}

PSID
get_world_sid ()
{
  static NO_COPY char world_sid_buf[40];
  static NO_COPY PSID world_sid = NULL;

  if (!world_sid)
    {
      world_sid = (PSID) world_sid_buf;
      get_ssid (world_sid, "S-1-1-0");
    }
  return world_sid;
}

int passwd_sem = 0;
int group_sem = 0;

int
get_id_from_sid (PSID psid, BOOL search_grp)
{
  if (!IsValidSid (psid))
    {
      __seterrno ();
      small_printf ("IsValidSid failed with %E");
      return -1;
    }

  /* First try to get SID from passwd or group entry */
  if (allow_ntsec)
    {
      char sidbuf[40];
      PSID sid = (PSID) sidbuf;
      int id = -1;

      if (! search_grp)
        {
          if (passwd_sem > 0)
            return 0;
          ++passwd_sem;

          struct passwd *pw;
          while ((pw = getpwent ()) != NULL)
            {
              if (get_pw_sid (sid, pw) && EqualSid (psid, sid))
                {
                  id = pw->pw_uid;
                  break;
                }
            }
          endpwent ();
          --passwd_sem;
          if (id >= 0)
            return id;
        }
      else
        {
          if (group_sem > 0)
            return 0;
          ++group_sem;

          struct group *gr;
          while ((gr = getgrent ()) != NULL)
            {
              if (get_gr_sid (sid, gr) && EqualSid (psid, sid))
                {
                  id = gr->gr_gid;
                  break;
                }
            }
          endgrent ();
          --group_sem;
          if (id >= 0)
            return id;
        }
    }
  /*
   * FIXME: I don't know if it really works in any case, but calling
   * LookupAccountSid() and NetxxxGetInfo() has very poor performance
   * especially in domain environments
   */

  /* It seems like the uid (or gid) is the last sub authority of the sid */
  int id = *GetSidSubAuthority(psid, *GetSidSubAuthorityCount(psid) - 1);

  if ((WORD) id == (WORD) -1)
    {
      /*
       * the last subauthority is sometimes -1, we have to go the
       * 'clean way' in this case
       */

      char account[100];
      char domain[100];
      DWORD account_length = 100;
      DWORD domain_length = 100;
      SID_NAME_USE account_type;

      if (!LookupAccountSid (NULL, psid, account, &account_length,
				domain, &domain_length, &account_type))
	{
	  __seterrno ();
	  return -1;
	}

      switch (account_type)
	{
	  case SidTypeUser:
	    {
	      struct passwd *pw = getpwnam (account);
	      return pw ? pw->pw_uid : getuid ();
	    }
	  case SidTypeGroup:
	    {
	      struct group *gr = getgrnam (account);
	      if (gr == NULL)
		 {
		   /* local groups are more like users than like groups */
		   struct passwd *pw = getpwnam (account);
		   return pw ? pw->pw_uid : getuid ();
		 }
	      return gr->gr_gid;
	    }
	  default:
	    return getuid ();
	}
    }

  return id;
}

static BOOL
legal_sid_type (SID_NAME_USE type)
{
  return type == SidTypeUser || type == SidTypeGroup
                 || SidTypeAlias || SidTypeWellKnownGroup;
}

BOOL
lookup_name (const char *name, const char *logsrv, PSID ret_sid)
{
  char sidbuf[40];
  PSID sid = (PSID) sidbuf;
  DWORD sidlen;
  char domuser[356];
  char dom[100];
  DWORD domlen;
  SID_NAME_USE acc_type;

  debug_printf ("name  : %s", name ? name : "NULL");

  if (! name)
    return FALSE;

  if (logsrv && *logsrv)
    {
      if (LookupAccountName (logsrv, name,
                             sid, (sidlen = 40, &sidlen),
                             dom, (domlen = 100, &domlen),
                             &acc_type)
          && legal_sid_type (acc_type))
        goto got_it;
      if (acc_type == SidTypeDomain)
        {
          strcat (strcat (strcpy (domuser, dom), "\\"), name);
          if (LookupAccountName (logsrv, domuser,
                                 sid, (sidlen = 40, &sidlen),
                                 dom, (domlen = 100, &domlen),
                                 &acc_type))
            goto got_it;
        }
    }
  if (LookupAccountName (NULL, name,
                         sid, (sidlen = 40, &sidlen),
                         dom, (domlen = 100, &domlen),
                         &acc_type)
      && legal_sid_type (acc_type))
    goto got_it;
  if (acc_type == SidTypeDomain)
    {
      strcat (strcat (strcpy (domuser, dom), "\\"), name);
      if (LookupAccountName (NULL, domuser,
                             sid, (sidlen = 40, &sidlen),
                             dom, (domlen = 100, &domlen),
                             &acc_type))
        goto got_it;
    }
  debug_printf ("LookupAccountName(%s) %E", name);
  __seterrno ();
  return FALSE;

got_it:
  debug_printf ("sid : [%d]", *GetSidSubAuthority((PSID) sid,
                              *GetSidSubAuthorityCount((PSID) sid) - 1));

  if (ret_sid)
    memcpy (ret_sid, sid, sidlen);

  return TRUE;
}

/* ReadSD reads a security descriptor from a file.
   In case of error, -1 is returned and errno is set.
   If the file doesn't have a SD, 0 is returned.
   Otherwise, the size of the SD is returned and
   the SD is copied to the buffer, pointed to by sdBuf.
   sdBufSize contains the size of the buffer. If
   it's too small, to contain the complete SD, 0 is
   returned and sdBufSize is set to the needed size
   of the buffer.
*/

LONG
ReadSD(const char *file, PSECURITY_DESCRIPTOR sdBuf, LPDWORD sdBufSize)
{
  /* Check parameters */
  if (! sdBufSize)
    {
      set_errno (EINVAL);
      return -1;
    }

  /* Open file for read */
  HANDLE hFile = CreateFile (file, GENERIC_READ,
			     FILE_SHARE_READ | FILE_SHARE_WRITE,
			     &sec_none_nih, OPEN_EXISTING,
			     FILE_ATTRIBUTE_NORMAL | FILE_FLAG_BACKUP_SEMANTICS,
			     NULL);

  if (hFile == INVALID_HANDLE_VALUE)
    {
      __seterrno ();
      return -1;
    }

  /* step through the backup streams and search for the security data */
  WIN32_STREAM_ID header;
  DWORD bytes_read = 0;
  LPVOID context = NULL;
  PSECURITY_DESCRIPTOR psd = NULL;
  DWORD datasize;
  LONG ret = 0;

  while (BackupRead (hFile, (LPBYTE) &header,
		     3 * sizeof (DWORD) + sizeof (LARGE_INTEGER),
		     &bytes_read, FALSE, TRUE, &context))
    {
      if (header.dwStreamId != BACKUP_SECURITY_DATA)
	continue;

      /* security data found */
      datasize = header.Size.u.LowPart + header.dwStreamNameSize;
      char b[datasize];

      if (! BackupRead (hFile, (LPBYTE) b, datasize, &bytes_read,
			FALSE, TRUE, &context))
	{
	  __seterrno ();
	  ret = -1;
	  break;
	}

      /* Check validity of the SD */
      psd = (PSECURITY_DESCRIPTOR) &b[header.dwStreamNameSize];
      if (! IsValidSecurityDescriptor (psd))
	continue;

      /* It's a valid SD */
      datasize -= header.dwStreamNameSize;
      debug_printf ("SD-Size: %d", datasize);

      /* buffer to small? */
      if (*sdBufSize < datasize)
	{
	  *sdBufSize = datasize;
	  ret = 0;
	  break;
	}

      if (sdBuf)
	memcpy (sdBuf, psd, datasize);

      ret = *sdBufSize = datasize;
      break;

    }
  BackupRead (hFile, NULL, 0, &bytes_read, TRUE, TRUE, &context);
  CloseHandle (hFile);
  return ret;
}

LONG
WriteSD(const char *file, PSECURITY_DESCRIPTOR sdBuf, DWORD sdBufSize)
{
  /* Check parameters */
  if (! sdBuf || ! sdBufSize)
    {
      set_errno (EINVAL);
      return -1;
    }

  HANDLE hFile = CreateFile (file,
			     WRITE_OWNER | WRITE_DAC,
                             FILE_SHARE_READ | FILE_SHARE_WRITE,
                             &sec_none_nih,
                             OPEN_EXISTING,
                             FILE_ATTRIBUTE_NORMAL | FILE_FLAG_BACKUP_SEMANTICS,
                             NULL);

  if (hFile == INVALID_HANDLE_VALUE)
    {
      __seterrno ();
      return -1;
    }

  LPVOID context = NULL;
  DWORD bytes_written = 0;
  WIN32_STREAM_ID header;

  memset (&header, 0, sizeof (header));
  /* write new security info header */
  header.dwStreamId = BACKUP_SECURITY_DATA;
  header.dwStreamAttributes = STREAM_CONTAINS_SECURITY;
  header.Size.u.HighPart = 0;
  header.Size.u.LowPart = sdBufSize;
  header.dwStreamNameSize = 0;
  if (!BackupWrite (hFile, (LPBYTE) &header,
		    3 * sizeof (DWORD) + sizeof (LARGE_INTEGER),
		    &bytes_written, FALSE, TRUE, &context))
    {
      __seterrno ();
      CloseHandle (hFile);
      return -1;
    }

  /* write new security descriptor */
  if (!BackupWrite (hFile, (LPBYTE) sdBuf,
		    header.Size.u.LowPart + header.dwStreamNameSize,
		    &bytes_written, FALSE, TRUE, &context))
    {
      /* Samba returns ERROR_NOT_SUPPORTED.
         FAT returns ERROR_INVALID_SECURITY_DESCR.
         This shouldn't return as error, but better be ignored. */
      DWORD ret = GetLastError ();
      if (ret != ERROR_NOT_SUPPORTED && ret != ERROR_INVALID_SECURITY_DESCR)
	{
	  __seterrno ();
	  BackupWrite (hFile, NULL, 0, &bytes_written, TRUE, TRUE, &context);
	  CloseHandle (hFile);
	  return -1;
	}
    }

  /* terminate the restore process */
  BackupWrite (hFile, NULL, 0, &bytes_written, TRUE, TRUE, &context);
  CloseHandle (hFile);
  return 0;
}

static int
set_process_privileges ()
{
  HANDLE hProcess = NULL;
  HANDLE hToken = NULL;
  LUID restore_priv;
  LUID backup_priv;
  char buf[sizeof (TOKEN_PRIVILEGES) + 2 * sizeof (LUID_AND_ATTRIBUTES)];
  TOKEN_PRIVILEGES *new_priv = (TOKEN_PRIVILEGES *) buf;
  int ret = -1;

  hProcess = OpenProcess (PROCESS_ALL_ACCESS, FALSE, GetCurrentProcessId ());
  if (! hProcess)
    {
      __seterrno ();
      goto out;
    }

  if (! OpenProcessToken (hProcess,
			  TOKEN_QUERY | TOKEN_ADJUST_PRIVILEGES,
			  &hToken))
    {
      __seterrno ();
      goto out;
    }

  if (! LookupPrivilegeValue (NULL, SE_RESTORE_NAME, &restore_priv))
    {
      __seterrno ();
      goto out;
    }
  if (! LookupPrivilegeValue (NULL, SE_BACKUP_NAME, &backup_priv))
    {
      __seterrno ();
      goto out;
    }

  new_priv->PrivilegeCount = 2;
  new_priv->Privileges[0].Luid = restore_priv;
  new_priv->Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
  new_priv->Privileges[1].Luid = backup_priv;
  new_priv->Privileges[1].Attributes = SE_PRIVILEGE_ENABLED;

  if (! AdjustTokenPrivileges (hToken, FALSE, new_priv, 0, NULL, NULL))
    {
      __seterrno ();
      goto out;
    }

  ret = 0;

  if (ret == -1)
    __seterrno ();

out:
  if (hToken)
    CloseHandle (hToken);
  if (hProcess)
    CloseHandle (hProcess);

  syscall_printf ("%d = set_process_privileges ()", ret);
  return ret;
}

static int
get_nt_attribute (const char *file, int *attribute)
{
  if (os_being_run != winNT)
    return 0;

  syscall_printf ("file: %s", file);

  if (set_process_privileges () < 0)
    return -1;

  /* Yeah, sounds too much, but I've seen SDs of 2100 bytes! */
  DWORD sd_size = 4096;
  char sd_buf[4096];
  PSECURITY_DESCRIPTOR sd = (PSECURITY_DESCRIPTOR) sd_buf;

  int ret;
  if ((ret = ReadSD (file, sd, &sd_size)) <= 0)
    {
      debug_printf ("ReadSD %E");
      return ret;
    }

  PSID owner_sid;
  PSID group_sid;
  BOOL dummy;

  if (! GetSecurityDescriptorOwner (sd, &owner_sid, &dummy))
    debug_printf ("GetSecurityDescriptorOwner %E");
  if (! GetSecurityDescriptorGroup (sd, &group_sid, &dummy))
    debug_printf ("GetSecurityDescriptorGroup %E");

  PACL acl;
  BOOL acl_exists;

  if (! GetSecurityDescriptorDacl (sd, &acl_exists, &acl, &dummy)
      || ! acl_exists
      || ! acl)
    {
      __seterrno ();
      debug_printf ("GetSecurityDescriptorDacl %E");
      return -1;
    }

  BOOL has_owner_bits = FALSE;
  BOOL has_group_bits = FALSE;
  BOOL has_world_bits = FALSE;

  for (DWORD i = 0; i < acl->AceCount; ++i)
    {
      ACCESS_ALLOWED_ACE *ace;

      if (GetAce (acl, i, (PVOID *) &ace))
	{
	  switch (ace->Header.AceType)
	    {
	    case ACCESS_ALLOWED_ACE_TYPE:
              if (owner_sid && EqualSid ((PSID) &ace->SidStart, owner_sid))
		{
		  *attribute &= ~S_IRWXU;
		  has_owner_bits = TRUE;
		  if (ace->Mask & FILE_READ_DATA)
		    *attribute |= S_IRUSR;
		  if (ace->Mask & FILE_WRITE_DATA)
		    *attribute |= S_IWUSR;
		  if (ace->Mask & FILE_EXECUTE)
		    *attribute |= S_IXUSR;
		}
              else if (group_sid && EqualSid ((PSID) &ace->SidStart, group_sid))
		{
		  *attribute &= ~S_IRWXG;
		  has_group_bits = TRUE;
		  if (ace->Mask & FILE_READ_DATA)
		    *attribute |= S_IRGRP;
		  if (ace->Mask & FILE_WRITE_DATA)
		    *attribute |= S_IWGRP;
		  if (ace->Mask & FILE_EXECUTE)
		    *attribute |= S_IXGRP;
		}
	      else if (EqualSid ((PSID) &ace->SidStart, get_world_sid ()))
		{
		  *attribute &= ~S_IRWXO;
		  has_world_bits = TRUE;
		  if (ace->Mask & FILE_READ_DATA)
		    {
		      *attribute |= S_IROTH;
		      if (! owner_sid || ! has_owner_bits)
			*attribute |= S_IRUSR;
		      if (! group_sid || ! has_group_bits)
			*attribute |= S_IRGRP;
		    }
		  if (ace->Mask & FILE_WRITE_DATA)
		    {
		      *attribute |= S_IWOTH;
		      if (! owner_sid || ! has_owner_bits)
			*attribute |= S_IWUSR;
		      if (! group_sid || ! has_group_bits)
			*attribute |= S_IWGRP;
		    }
		  if (ace->Mask & FILE_EXECUTE)
		    {
		      *attribute |= S_IXOTH;
		      if (! owner_sid || ! has_owner_bits)
			*attribute |= S_IXUSR;
		      if (! group_sid || ! has_group_bits)
			*attribute |= S_IXGRP;
		      // Sticky bit for directories according to linux rules.
		      // No sense for files.
		      if (! (ace->Mask & FILE_DELETE_CHILD) &&
			  S_ISDIR(*attribute))
			*attribute |= S_ISVTX;
		    }
		}
	      break;
	    case ACCESS_DENIED_ACE_TYPE:
	      // Still ignored!
	      break;
	    default:
	      break;
	    }
	}
    }
  syscall_printf ("file: %s %x", file, *attribute);
  return 0;
}

int
get_file_attribute (int use_ntsec, const char *file, int *attribute)
{
  if (!attribute)
    {
      set_errno (EINVAL);
      return -1;
    }

  int res;

  if (use_ntsec && allow_ntsec)
    {
      res = get_nt_attribute (file, attribute);
      if (!res)
        return 0;
    }

  res = NTReadEA (file, ".UNIXATTR", (char *) attribute, sizeof (*attribute));

  // symlinks are anything for everyone!
  if ((*attribute & S_IFLNK) == S_IFLNK)
    *attribute |= S_IRWXU | S_IRWXG | S_IRWXO;

  if (res > 0)
    return 0;
  set_errno (ENOSYS);
  return -1;
}

PSECURITY_DESCRIPTOR
alloc_sd (uid_t uid, gid_t gid, const char *logsrv, int attribute,
	  PSECURITY_DESCRIPTOR sd_ret, DWORD *sd_size_ret)
{
  if (os_being_run != winNT)
    return NULL;

  if (! sd_ret || ! sd_size_ret)
    {
      set_errno (EINVAL);
      return NULL;
    }

  char user[256];
  char *owner_sid_buf[40]; // SIDs are max 40 bytes long
  PSID owner_sid = NULL;
  char *group_sid_buf[40]; // SIDs are max 40 bytes long
  PSID group_sid = NULL;

  SECURITY_DESCRIPTOR sd;
  PSECURITY_DESCRIPTOR psd = NULL;
  DWORD sd_size = 0;

  memset (&sd, 0, sizeof (sd));

  size_t acl_len;
  char acl_buf[256];
  PACL acl = (PACL) acl_buf;

  struct passwd *pw = getpwuid (uid);
  strcpy (user, pw ? pw->pw_name : getlogin ());
  struct group *grp = getgrgid (gid);

  owner_sid = (PSID) owner_sid_buf;
  if ((! pw || ! get_pw_sid (owner_sid, pw))
      && ! lookup_name (user, logsrv, owner_sid))
    goto done;

  debug_printf ("user: %s [%d]", user,
                *GetSidSubAuthority((PSID) owner_sid,
		*GetSidSubAuthorityCount((PSID) owner_sid) - 1));

  if (grp)
    {
      group_sid = (PSID) group_sid_buf;
      if ((! grp || ! get_gr_sid (group_sid, grp))
          && ! lookup_name (grp->gr_name, logsrv, group_sid))
	goto done;
    }
  else
    debug_printf ("no group");

  if (! InitializeSecurityDescriptor (&sd, SECURITY_DESCRIPTOR_REVISION))
    debug_printf ("InitializeSecurityDescriptor  %E");
  if (! SetSecurityDescriptorOwner(&sd, owner_sid, FALSE))
    {
      __seterrno ();
      goto done;
    }
  if (group_sid
      && ! SetSecurityDescriptorGroup(&sd, group_sid, FALSE))
    {
      __seterrno ();
      goto done;
    }

  acl_len = sizeof (ACL);

  acl_len += sizeof (ACCESS_ALLOWED_ACE) - sizeof (DWORD)
             + GetLengthSid (owner_sid);

  if (! group_sid || ! EqualSid (group_sid, get_admin_sid ()))
    acl_len += sizeof (ACCESS_ALLOWED_ACE) - sizeof (DWORD)
	       + GetLengthSid (get_admin_sid ());

  if (group_sid && ((attribute & S_IRWXG)
                    || EqualSid (group_sid, get_admin_sid ())))
    acl_len += sizeof (ACCESS_ALLOWED_ACE) - sizeof (DWORD)
	       + GetLengthSid (group_sid);

  if (attribute & S_IRWXO)
    acl_len += sizeof (ACCESS_ALLOWED_ACE) - sizeof (DWORD)
	       + GetLengthSid (get_world_sid ());

  debug_printf ("ACL-Size: %d", acl_len);

  if (! InitializeAcl (acl, acl_len, ACL_REVISION))
    debug_printf ("InitializeAcl %E");

  if (attribute != -1)
    {
      int ace_off = 0;
      ACCESS_ALLOWED_ACE *ace;

      DWORD access = STANDARD_RIGHTS_ALL;

      /*
       * VTX bit may only be set if executable for `other' is set.
       * For correct handling under WinNT, FILE_DELETE_CHILD has to
       * be (un)set in each ACE.
      */
      if (! (attribute & S_IXOTH))
        attribute &= ~S_ISVTX;

      if (attribute & S_IRUSR)
        access |= FILE_GENERIC_READ | FILE_WRITE_ATTRIBUTES | FILE_WRITE_EA;
      if (attribute & S_IWUSR)
        access |= FILE_GENERIC_WRITE | DELETE;
      if (attribute & S_IXUSR)
        access |= FILE_GENERIC_EXECUTE;
      if (! (attribute & S_ISVTX))
        access |= FILE_DELETE_CHILD;
      if (! AddAccessAllowedAce (acl, ACL_REVISION, access, owner_sid))
        debug_printf ("AddAccessAllowedAce(owner) %E");
      if (GetAce(acl, 0, (PVOID *) &ace))
        ace->Header.AceFlags |= OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE;

      if (! group_sid || ! EqualSid (group_sid, get_admin_sid ()))
	{
	  if (! AddAccessAllowedAce (acl, ACL_REVISION,
	                             STANDARD_RIGHTS_READ | WRITE_OWNER,
				     get_admin_sid ()))
	    debug_printf ("AddAccessAllowedAce(admin) %E");
	  if (GetAce(acl, 1, (PVOID *) &ace))
	    ace->Header.AceFlags |= OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE;
	  ace_off = 1;
	}

      if (group_sid && ((attribute & S_IRWXG)
                        || EqualSid (group_sid, get_admin_sid ())))
	{
	  access = STANDARD_RIGHTS_READ;
          if (EqualSid (group_sid, get_admin_sid ()))
	    access |= WRITE_DAC | WRITE_OWNER
		      | FILE_WRITE_ATTRIBUTES | FILE_WRITE_EA;
	  if (attribute & S_IRGRP)
	    access |= FILE_GENERIC_READ;
	  if (attribute & S_IWGRP)
	    access |= STANDARD_RIGHTS_ALL | FILE_GENERIC_WRITE | DELETE;
	  if (attribute & S_IXGRP)
	    access |= FILE_GENERIC_EXECUTE;
          if (! (attribute & S_ISVTX))
            access |= FILE_DELETE_CHILD;
	  if (! AddAccessAllowedAce (acl, ACL_REVISION, access, group_sid))
	    debug_printf ("AddAccessAllowedAce(group) %E");
	  if (GetAce(acl, 1 + ace_off, (PVOID *) &ace))
	    ace->Header.AceFlags |= OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE;
	}

      if (attribute & S_IRWXO)
        {
          access = STANDARD_RIGHTS_READ;
          if (attribute & S_IROTH)
            access |= FILE_GENERIC_READ;
          if (attribute & S_IWOTH)
            access |= STANDARD_RIGHTS_ALL | FILE_GENERIC_WRITE | DELETE;
          if (attribute & S_IXOTH)
            access |= FILE_GENERIC_EXECUTE;
          if (! (attribute & S_ISVTX))
            access |= FILE_DELETE_CHILD;

          if (! AddAccessAllowedAce (acl, ACL_REVISION, access, get_world_sid ()))
            debug_printf ("AddAccessAllowedAce(world) %E");
          if (GetAce(acl, 2 + ace_off, (PVOID *) &ace))
            ace->Header.AceFlags |= OBJECT_INHERIT_ACE | CONTAINER_INHERIT_ACE;
        }

      if (! SetSecurityDescriptorDacl (&sd, TRUE, acl, FALSE))
	debug_printf ("SetSecurityDescriptorDacl %E");
    }

  sd_size = *sd_size_ret;

  memset (sd_ret, 0, sd_size);
  if (! MakeSelfRelativeSD (&sd, sd_ret, &sd_size))
    {
      __seterrno ();
      goto done;
    }
  psd = sd_ret;
  debug_printf ("Created SD-Size: %d", *sd_size_ret);

done:
  *sd_size_ret = sd_size;
  return psd;
}

static int
set_nt_attribute (const char *file, uid_t uid, gid_t gid,
                  const char *logsrv, int attribute)
{
  if (os_being_run != winNT)
    return 0;

  if (set_process_privileges () < 0)
    return -1;

  DWORD sd_size = 512;
  char sd_buf[512];
  PSECURITY_DESCRIPTOR psd = (PSECURITY_DESCRIPTOR) sd_buf;

  if (! (psd = alloc_sd (uid, gid, logsrv, attribute, psd, &sd_size)))
    return -1;

  int ret = WriteSD (file, psd, sd_size);

  return ret;
}

int
set_file_attribute (int use_ntsec, const char *file,
                    uid_t uid, gid_t gid,
                    int attribute, const char *logsrv)
{
  // symlinks are anything for everyone!
  if ((attribute & S_IFLNK) == S_IFLNK)
    attribute |= S_IRWXU | S_IRWXG | S_IRWXO;

  BOOL ret = NTWriteEA (file, ".UNIXATTR",
			(char *) &attribute, sizeof (attribute));
  if (!use_ntsec || !allow_ntsec)
    {
      if (! ret)
	{
	  __seterrno ();
	  return -1;
	}
      return 0;
    }

  int ret2 = set_nt_attribute (file, uid, gid, logsrv, attribute);
  syscall_printf ("%d = set_file_attribute (%s, %d, %d, %p)",
		  ret2, file, uid, gid, attribute);
  return ret2;
}

int
set_file_attribute (int use_ntsec, const char *file, int attribute)
{
  return set_file_attribute (use_ntsec, file,
                             myself->uid, myself->gid,
                             attribute, myself->logsrv);
}

